#use "ass12.ml"

(* Perform all possible beta-reductions on an expression until the normal form
 * has been reached *)
let rec eval exp = match exp with
      MonopAp (op, e) -> MonopAp (op, eval e)
    | BinopAp (op, e1, e2) -> BinopAp (op, eval e1, eval e2)
    | Fun (x, e) -> Fun (x, eval e)
    | FunAp (Fun (x, Var y), a) when y = x -> eval a (* identity function *)
    | FunAp (Fun (y, e), a) -> (* prevent recursion *)
        let result = (subs e y a) in
        if result = exp then result else eval result
    | FunAp (e1, e2) -> (* prevent recursion *)
        let result = FunAp (eval e1, eval e2) in
        if result = exp then result else eval result
    | Cond (cond, e1, e2) -> Cond (eval cond, eval e1, eval e2)
    | Let (x, e1, e2) -> eval (subs e2 x e1)
    | LetRec (x, e1, e2) -> LetRec (x, eval e1, eval e2)
    | _ -> exp
